#!/usr/local/bin/python

"""module assem:
    provides an assembler and disassembler for compiled Python code objects.

    Large junks of this file were stolen from the standard Python library dis.py.

    Author: Mike McDonald <mikemac@sgi.com>

    History:
        Version: 1.0 April 8, 1997 - first released
"""

import sys, types, string, new

class disassemble:
    """class disassemble:
    provides an interface for disassembling and reassembling Python code objects.

    constructor: disassemble(co)
        co must be a code object (types.CodeType)

    instance variable: instructions
        a list of Python instructions. This list can be modified before calling assemble().

    method: assemble()
        reassembles the code object from the instructions and returns it.

    method: targeted_labels()
        returns a list of labels that are the targets of jump instructions.

    method: dump(fp=stdout)
        prints a human readable printout of the instructions to the file pointer, fp.
"""

    def __init__(self, co):
	if type(co) == types.FunctionType:
	    co = co.func_code
	elif type(co) != types.CodeType:
	    raise TypeError, "The object '%s' is not a code object (types.CodeType)" % co
	self.args = co.co_varnames[:co.co_argcount]
	self.doc_string = co.co_consts[0]
	self.flags = co.co_flags
	self.name = co.co_name
	self.filename = co.co_filename
	code = co.co_code
	n = len(code)
	i = 0
	self.instructions = []
	while i < n:
	    c = code[i]
	    op = ord(c)
	    instruction_number = i
	    i = i+1
	    oparg = None
	    if op >= HAVE_ARGUMENT:
		oparg = ord(code[i]) + ord(code[i+1])*256
		i = i+2
		if op in hasconst:
		    oparg = co.co_consts[oparg]
		elif op in hasname:
		    oparg = co.co_names[oparg]
		elif op in hasjrel:
		    oparg = i + oparg
		elif op in haslocal:
		    oparg = co.co_varnames[oparg]
	    instruction = [instruction_number, opname[op], oparg]
	    self.instructions.append(instruction)

    def __repr__(self):
	return "<disassembled code '%s'>" % self.name

    def compute_instruction_offsets(self):
	i_num = 0
	self.offsets = {}
	for instr in self.instructions:
	    label = instr[0]
	    opname = instr[1]
	    self.offsets[label] = i_num
	    i_num = i_num + 1
	    if opcode[opname] >= HAVE_ARGUMENT:
		i_num = i_num + 2

    def find_address(self, label):
	if self.offsets.has_key(label):
	    return self.offsets[label]
	else:
	    keys = self.offsets.keys()
	    keys.sort()
	    for i in keys:
		if i > label:
		    return self.offsets[i]
	    raise ValueError, "Cannot find an address to the label '%s' in the list of labels %s" % (label, keys)

    def lookup(self, l, item):
	if not item in l:
	    l.append(item)
	return l.index(item)

    def assemble(self):
	"""assemble() returns a code object contructed from the instructions."""
	code = ''
	i_num = 0
	offsets = self.compute_instruction_offsets()
	consts = [self.doc_string]
	names = []
	varnames = list(self.args)
	for instr in self.instructions:
	    label = instr[0]
	    opname = instr[1]
	    arg = instr[2]
	    op = opcode[opname]
	    i_num = i_num + 1
	    if op >= HAVE_ARGUMENT:
		oparg = 0
		i_num = i_num + 2
		if op in hasconst:
		    oparg = self.lookup(consts, arg)
		elif op in hasname:
		    oparg = self.lookup(names, arg)
		elif op in hasjrel:
		    oparg = self.find_address(arg) - i_num
		elif op in hasjabs:
		    oparg = self.find_address(arg)
		elif op in haslocal:
		    self.lookup(names, arg)	# varnames are also on the names list
		    oparg = self.lookup(varnames, arg)
		else:
		    oparg = arg
		code = code + '%c%c%c' % (op, oparg % 256, oparg / 256)
	    else:
		code = code + '%c' % (op, )
	return new.code(len(self.args), len(varnames), self.flags,
			code,
			tuple(consts), tuple(names),
			tuple(varnames), self.filename, self.name)

    def targeted_labels(self):
	jump_targets = []
	for instr in self.instructions:
	    opname = instr[1]
	    arg = instr[2]
	    op = opcode[opname]
	    if op in hasjabs or op in hasjrel:
		jump_targets.append(arg)
	return jump_targets
    
    def dump(self, fp=sys.stdout):
	used_labels = self.targeted_labels()
	fp.write("# %s%s\n" % (self.name, self.args))
	for instr in self.instructions:
	    label = instr[0]
	    opname = instr[1]
	    arg = instr[2]
	    op = opcode[opname]
	    if label in used_labels:
		fp.write("%8d " % (label, ))
	    else:
		fp.write("\t ")
	    fp.write("%-16s " % (opname, ))
	    if op >= HAVE_ARGUMENT:
		fp.write("%s" % (arg, ))
	    fp.write("\n")
	

hasconst = []
hasname = []
hasjrel = []
hasjabs = []
haslocal = []

opname = [''] * 256
opcode = {}
for op in range(256):
    opname[op] = '<' + `op` + '>'
    opcode[opname[op]] = op

def def_op(name, op):
    opname[op] = name
    opcode[name] = op

def name_op(name, op):
    def_op(name, op)
    hasname.append(op)

def jrel_op(name, op):
    def_op(name, op)
    hasjrel.append(op)

def jabs_op(name, op):
    def_op(name, op)
    hasjabs.append(op)

# Instruction opcodes for compiled code

def_op('STOP_CODE', 0)
def_op('POP_TOP', 1)
def_op('ROT_TWO', 2)
def_op('ROT_THREE', 3)
def_op('DUP_TOP', 4)

def_op('UNARY_POSITIVE', 10)
def_op('UNARY_NEGATIVE', 11)
def_op('UNARY_NOT', 12)
def_op('UNARY_CONVERT', 13)
def_op('UNARY_CALL', 14)
def_op('UNARY_INVERT', 15)

def_op('BINARY_POWER', 19)
def_op('BINARY_MULTIPLY', 20)
def_op('BINARY_DIVIDE', 21)
def_op('BINARY_MODULO', 22)
def_op('BINARY_ADD', 23)
def_op('BINARY_SUBTRACT', 24)
def_op('BINARY_SUBSCR', 25)
def_op('BINARY_CALL', 26)

def_op('SLICE+0', 30)
def_op('SLICE+1', 31)
def_op('SLICE+2', 32)
def_op('SLICE+3', 33)

def_op('STORE_SLICE+0', 40)
def_op('STORE_SLICE+1', 41)
def_op('STORE_SLICE+2', 42)
def_op('STORE_SLICE+3', 43)

def_op('DELETE_SLICE+0', 50)
def_op('DELETE_SLICE+1', 51)
def_op('DELETE_SLICE+2', 52)
def_op('DELETE_SLICE+3', 53)

def_op('STORE_SUBSCR', 60)
def_op('DELETE_SUBSCR', 61)

def_op('PRINT_EXPR', 70)
def_op('PRINT_ITEM', 71)
def_op('PRINT_NEWLINE', 72)

def_op('BREAK_LOOP', 80)
def_op('RAISE_EXCEPTION', 81)
def_op('LOAD_LOCALS', 82)
def_op('RETURN_VALUE', 83)
def_op('EXEC_STMT', 85)
def_op('BUILD_FUNCTION', 86)
def_op('POP_BLOCK', 87)
def_op('END_FINALLY', 88)
def_op('BUILD_CLASS', 89)

HAVE_ARGUMENT = 90		# Opcodes from here have an argument: 

name_op('STORE_NAME', 90)	# Index in name list 
name_op('DELETE_NAME', 91)	# "" 
def_op('UNPACK_TUPLE', 92)	# Number of tuple items 
def_op('UNPACK_LIST', 93)	# Number of list items 
def_op('UNPACK_ARG', 94)	# Number of arguments expected
name_op('STORE_ATTR', 95)	# Index in name list 
name_op('DELETE_ATTR', 96)	# ""
name_op('STORE_GLOBAL', 97)	# ""
name_op('DELETE_GLOBAL', 98)	# ""
name_op('UNPACK_VARARG', 99)	# Minimal number of arguments

def_op('LOAD_CONST', 100)	# Index in const list 
hasconst.append(100)
name_op('LOAD_NAME', 101)	# Index in name list 
def_op('BUILD_TUPLE', 102)	# Number of tuple items 
def_op('BUILD_LIST', 103)	# Number of list items 
def_op('BUILD_MAP', 104)	# Always zero for now 
name_op('LOAD_ATTR', 105)	# Index in name list 
def_op('COMPARE_OP', 106)	# Comparison operator 
name_op('IMPORT_NAME', 107)	# Index in name list 
name_op('IMPORT_FROM', 108)	# Index in name list 

jrel_op('JUMP_FORWARD', 110)	# Number of bytes to skip 
jrel_op('JUMP_IF_FALSE', 111)	# "" 
jrel_op('JUMP_IF_TRUE', 112)	# "" 
jabs_op('JUMP_ABSOLUTE', 113)	# Target byte offset from beginning of code 
jrel_op('FOR_LOOP', 114)	# Number of bytes to skip 

name_op('LOAD_LOCAL', 115)	# Index in name list
name_op('LOAD_GLOBAL', 116)	# Index in name list
def_op('SET_FUNC_ARGS', 117)	# Argcount

jrel_op('SETUP_LOOP', 120)	# Distance to target address
jrel_op('SETUP_EXCEPT', 121)	# ""
jrel_op('SETUP_FINALLY', 122)	# ""

def_op('RESERVE_FAST', 123)	# Number of local variables
hasconst.append(123)
def_op('LOAD_FAST', 124)	# Local variable number
haslocal.append(124)
def_op('STORE_FAST', 125)	# Local variable number
haslocal.append(125)
def_op('DELETE_FAST', 126)	# Local variable number
haslocal.append(126)

def_op('SET_LINENO', 127)	# Current line number
SET_LINENO = 127

def_op('RAISE_VARARGS', 130)
def_op('CALL_FUNCTION', 131)
def_op('MAKE_FUNCTION', 132)
def_op('BUILD_SLICE', 133)

if __name__ == '__main__':
    import dis
    func = dis.disassemble.func_code
    c = disassemble(func)
    co = c.assemble()
    if co == func:
	print "Assem self test passed"
    else:
	print "Assem self test failed!!!"
	print "co.co_names == orig_func.co_names?", co.co_names == func.co_names
	print "co.co_varnames == orig_func.co_varnames?", co.co_varnames == func.co_varnames
	print "co.co_consts == orig_func.co_consts?", co.co_consts == func.co_consts
