#!/usr/local/bin/python

"""module ctransform : 
    provides an interface to load, transform, and
    save compiled module files ('.pyc'). A code transformation function
    takes one argument, a code object to be transformed and returns that 
    new code object.

    Author: Mike McDonald <mikemac@sgi.com>

    History:
        Version: 1.0 April 8, 1997 - first released
"""

import sys, types, assem, new, marshal, os, ospath

class compiled_module:
    """class compiled_module:
    encapsulates the contents of a compile module's save file ('.pyc').

    constructor: compiled_module(file, header, code_object)
        is not intended to be instantiated by the user.
	Use load_compiled_module() instead.

    class variable: backup_ext 
        specifies what filename extension to use
	when backing up the original source file. The default value is '.org'.
	Setting backup_ext to None causes no backup file to be generated. If
	a backup file with the specified extension already exists, a new one
	will not be created.

    method: dump()
        saves the compiled module back out to the save file,
	possibly backing up the previous version.
"""

    backup_ext = '.org'

    def __init__(self, file, header, code_object):
	self.file = file
	self.header = header
	self.code_object = code_object

    def dump(self):
	"""saves the compiled module into file, possibly backing it up first."""
	(base, ext) = os.path.splitext(self.file)
	if not os.path.exists(base + self.backup_ext):
	    os.rename(self.file, base + self.backup_ext)
	fp = open(self.file, 'wb')
	fp.write(self.header)
	marshal.dump(self.code_object, fp)
	fp.close()

def load_compiled_module(path):
    """load_compiled_module(path) creates an instance of compiled_module
    from the contents of the file 'path'."""
    (base, ext) = os.path.splitext(path)
    if ext != '.pyc':
	raise ValueError, "The file '" + path + "' is not a compiled Python file."
	return None
    fp = open(path, 'rb')
    if not fp:
	return None
    header = fp.read(8)
    co = marshal.load(fp)
    fp.close()
    return compiled_module(path, header, co)

def copy_function(co):
    """copy_function(co) is the default transformation function for transform_compiled_module().
    It simply returns an identical copy of the argument code object."""
    code = assem.disassemble(co)
    return code.assemble()

def __transform_co(co, transform_func = copy_function):
    """transform_co(co, transform_func = copy_function) applies the
    transform_func to the co_code and recursively calls itsself on
    all code_objects in the co_consts field. It returns a new code_object.
    """
    new_consts = ()
    for const in co.co_consts:
	if type(const) == types.CodeType:
	    const = __transform_co(const, transform_func)
	new_consts = new_consts + (const, )
    return transform_func(new.code(co.co_argcount, co.co_nlocals, co.co_flags, co.co_code,
				   new_consts,
				   co.co_names, co.co_varnames, co.co_filename, co.co_name))

def transform_compiled_module(old_module, transform_func = copy_function):
    """transform_compiled_module(old_module, transform_func = copy_function) applies the
    transform_func to each code object in the compiled_module instance, old_module. It returns
    a new instance of the class compiled_module."""
    co = old_module.code_object
    return compiled_module(old_module.file, old_module.header,
			   __transform_co(co, transform_func))

if __name__ == '__main__':
    mod = load_compiled_module("/usr/local/lib/python1.4/dis.pyc")
    new_mod = transform_compiled_module(mod, copy_function)
    new_mod.dump()
    print 'diffing /usr/local/lib/python1.4/dis.pyc /usr/local/lib/python1.4/dis' + new_mod.backup_ext + ' (should be no difference)'
    os.system('diff /usr/local/lib/python1.4/dis.pyc /usr/local/lib/python1.4/dis' + new_mod.backup_ext)
    print 'done'
    os.rename('/usr/local/lib/python1.4/dis' + new_mod.backup_ext, '/usr/local/lib/python1.4/dis.pyc')
